""" The onegov town collection of files uploaded to the site. """

import base64
import morepath

from onegov.core.filestorage import random_filename
from onegov.core.security import Private
from onegov.town import _
from onegov.town.app import TownApp
from onegov.town.elements import Link
from onegov.town.layout import DefaultLayout
from onegov.town.models import FileCollection
from webob import exc


@TownApp.html(model=FileCollection, template='files.pt', permission=Private)
def view_get_file_collection(self, request):
    request.include('dropzone')

    files = [
        Link(text=file_.original_name, url=request.link(file_))
        for file_ in self.files
    ]

    layout = DefaultLayout(self, request)
    layout.breadcrumbs = [
        Link(_("Homepage"), layout.homepage_url),
        Link(_("Files"), request.link(self))
    ]

    return {
        'layout': layout,
        'title': _('Files'),
        'files': files,
    }


@TownApp.json(model=FileCollection, permission=Private, name='json')
def view_get_file_collection_json(self, request):
    return [
        {
            'link': request.link(file_),
            'title': file_.original_name.decode('utf-8')
        }
        for file_ in self.files
    ]


def handle_file_upload(self, request):
    """ Stores the file given with the request and returns the url to the
    resulting file.

    """

    extension = request.params['file'].filename.split('.')[-1]
    name = request.params['file'].filename.encode('utf-8')
    # Pad with whitespace to avoid '=' padding in base64 encoded value.
    if len(name) % 3 != 0:
        name = name.ljust(len(name) + (3 - len(name) % 3))
    name = base64.urlsafe_b64encode(name).decode('utf-8')

    filename = '{}-{}.{}'.format(name, random_filename(), extension)

    return self.store_file(request.params['file'].file, filename)


@TownApp.view(model=FileCollection, name='upload', request_method='POST',
              permission=Private)
def view_upload_file(self, request):
    request.assert_valid_csrf_token()

    handle_file_upload(self, request)

    return morepath.redirect(request.link(self))


@TownApp.json(model=FileCollection, name='upload.json', request_method='POST',
              permission=Private)
def view_upload_file_by_json(self, request):
    request.assert_valid_csrf_token()

    try:
        file_ = handle_file_upload(self, request)
    except exc.HTTPUnsupportedMediaType:
        return {
            'error': True,
            'message': request.translate(_("This file type is not supported"))
        }
    except exc.HTTPRequestHeaderFieldsTooLarge:
        return {
            'error': True,
            'message': request.translate(_("The file name is too long"))
        }

    return {
        'filelink': request.link(file_),
        'filename': file_.original_name.decode('utf-8'),
    }



import transaction

from base64 import b64decode
from datetime import datetime
from freezegun import freeze_time
from onegov.ticket import Handler, Ticket, TicketCollection
from onegov.user import UserCollection
from webtest import TestApp as Client
from sedate import ensure_timezone


def get_cronjob_by_name(app, name):
    for cronjob in app.config.cronjob_registry.cronjobs.values():
        if name in cronjob.name:
            return cronjob


def get_cronjob_url(cronjob):
    return '/cronjobs/{}'.format(cronjob.id)


def register_echo_handler(handlers):

    class EchoTicket(Ticket):
        __mapper_args__ = {'polymorphic_identity': 'ECO'}
        es_type_name = 'eco_tickets'

    @handlers.registered_handler('ECO')
    class EchoHandler(Handler):

        @property
        def deleted(self):
            return False

        @property
        def email(self):
            return self.data.get('email')

        @property
        def title(self):
            return self.data.get('title')

        @property
        def subtitle(self):
            return self.data.get('subtitle')

        @property
        def group(self):
            return self.data.get('group')

        def get_summary(self, request):
            return self.data.get('summary')

        def get_links(self, request):
            return self.data.get('links')


def get_mail(outbox, index):
    message = outbox[0]

    return {
        'from': message['From'],
        'reply_to': message['Reply-To'],
        'subject': message['Subject'],
        'text': b64decode(
            ''.join(message.get_payload(0).as_string().splitlines()[3:])
        ).decode('utf-8'),
        'html': b64decode(
            ''.join(message.get_payload(1).as_string().splitlines()[3:])
        ).decode('utf-8')
    }


def test_ticket_statistics(town_app, smtp, handlers):
    register_echo_handler(handlers)

    client = Client(town_app)

    job = get_cronjob_by_name(town_app, 'ticket_statistics')
    job.app = town_app

    url = get_cronjob_url(job)

    tz = ensure_timezone('Europe/Zurich')

    assert len(smtp.outbox) == 0

    # do not run on the weekends
    with freeze_time(datetime(2016, 1, 2, tzinfo=tz)):
        client.get(url)

    with freeze_time(datetime(2016, 1, 3, tzinfo=tz)):
        client.get(url)

    assert len(smtp.outbox) == 0

    session = town_app.session()
    collection = TicketCollection(session)

    tickets = [
        collection.open_ticket(
            handler_id='1',
            handler_code='ECO',
            title="Title",
            group="Group",
            email="citizen@example.org",
            created=datetime(2016, 1, 2, 10, tzinfo=tz),
        ),
        collection.open_ticket(
            handler_id='2',
            handler_code='ECO',
            title="Title",
            group="Group",
            email="citizen@example.org",
            created=datetime(2016, 1, 2, 10, tzinfo=tz)
        ),
        collection.open_ticket(
            handler_id='3',
            handler_code='ECO',
            title="Title",
            group="Group",
            email="citizen@example.org",
            created=datetime(2016, 1, 2, 10, tzinfo=tz)
        ),
        collection.open_ticket(
            handler_id='4',
            handler_code='ECO',
            title="Title",
            group="Group",
            email="citizen@example.org",
            created=datetime(2016, 1, 2, 10, tzinfo=tz)
        ),
        collection.open_ticket(
            handler_id='5',
            handler_code='ECO',
            title="Title",
            group="Group",
            email="citizen@example.org",
            created=datetime(2016, 1, 2, 10, tzinfo=tz)
        ),
        collection.open_ticket(
            handler_id='6',
            handler_code='ECO',
            title="Title",
            group="Group",
            email="citizen@example.org",
            created=datetime(2016, 1, 2, 10, tzinfo=tz)
        )
    ]

    users = UserCollection(session).query().all()
    user = users[0]

    users[1].data = {'daily_ticket_statistics': False}

    for ticket in tickets:
        ticket.created = datetime(2016, 1, 2, 10, tzinfo=tz)

    for pending in tickets[1:3]:
        pending.accept_ticket(user)
        pending.modified = datetime(2016, 1, 2, 10, tzinfo=tz)

    for closed in tickets[3:6]:
        closed.accept_ticket(user)
        closed.close_ticket()
        closed.modified = datetime(2016, 1, 2, 10, tzinfo=tz)

    transaction.commit()

    with freeze_time(datetime(2016, 1, 4, tzinfo=tz)):
        client.get(url)

    assert len(smtp.outbox) == 1
    message = get_mail(smtp.outbox, 0)

    assert message['subject'] == 'Govikon OneGov Cloud Status'
    txt = message['text']
    assert "Folgendes ist während des Wochenendes auf der Govikon" in txt
    assert "6 Tickets wurden eröffnet." in txt
    assert "2 Tickets wurden angenommen." in txt
    assert "3 Tickets wurden geschlossen." in txt
    assert "Zur Zeit ist 1 Ticket offen" in txt
    assert "2 Tickets sind in Bearbeitung" in txt
    assert "Wir wünschen Ihnen eine schöne Woche!" in txt
    assert "Das OneGov Cloud Team" in txt
    assert "/unsubscribe?token=" in txt
    assert "abmelden" in txt



from onegov.core.security import Private
from onegov.town import _, TownApp
from onegov.town.models import SiteCollection


@TownApp.json(model=SiteCollection, permission=Private)
def get_site_collection(self, request):
    """ Returns a list of internal links to be used by the redactor.

    See `<http://imperavi.com/redactor/plugins/predefined-links/>`_

    """

    objects = self.get()

    groups = [
        ('topics', request.translate(_("Topics"))),
        ('news', request.translate(_("Latest news"))),
        ('forms', request.translate(_("Forms"))),
        ('resources', request.translate(_("Resources"))),
    ]

    links = []

    for id, label in groups:
        for obj in objects[id]:
            # in addition to the default url/name pairings we use a group
            # label which will be used as optgroup label
            links.append({
                'group': label,
                'name': obj.title,
                'url': request.link(obj)
            })

    return links



from onegov.town.forms.allocation import (
    DaypassAllocationForm,
    DaypassAllocationEditForm,
    RoomAllocationForm,
    RoomAllocationEditForm
)
from onegov.town.forms.event import EventForm
from onegov.town.forms.form_definition import (
    BuiltinDefinitionForm,
    CustomDefinitionForm
)
from onegov.town.forms.newsletter import NewsletterForm, NewsletterSendForm
from onegov.town.forms.page import LinkForm, PageForm
from onegov.town.forms.person import PersonForm
from onegov.town.forms.reset_password import (
    RequestPasswordResetForm,
    PasswordResetForm
)
from onegov.town.forms.reservation import ReservationForm
from onegov.town.forms.resource import ResourceForm, ResourceCleanupForm
from onegov.town.forms.settings import SettingsForm
from onegov.town.forms.signup import SignupForm
from onegov.town.forms.userprofile import UserProfileForm

__all__ = [
    'BuiltinDefinitionForm',
    'CustomDefinitionForm',
    'DaypassAllocationForm',
    'DaypassAllocationEditForm',
    'EventForm',
    'LinkForm',
    'NewsletterForm',
    'NewsletterSendForm',
    'PageForm',
    'PasswordResetForm',
    'PersonForm',
    'ReservationForm',
    'ResourceForm',
    'ResourceCleanupForm',
    'RequestPasswordResetForm',
    'RoomAllocationForm',
    'RoomAllocationEditForm',
    'SettingsForm',
    'SignupForm',
    'UserProfileForm'
]



import morepath

from collections import OrderedDict
from itertools import groupby
from onegov.core.security import Public, Private
from onegov.libres import ResourceCollection
from onegov.libres.models import Resource
from onegov.town import TownApp, _
from onegov.town import utils
from onegov.town.elements import Link
from onegov.town.forms import ResourceForm, ResourceCleanupForm
from onegov.town.layout import ResourcesLayout, ResourceLayout
from onegov.town.models.resource import DaypassResource, RoomResource
from sqlalchemy.sql.expression import nullsfirst
from webob import exc


RESOURCE_TYPES = {
    'daypass': {
        'success': _("Added a new daypass"),
        'title': _("New daypass"),
        'class': DaypassResource
    },
    'room': {
        'success': _("Added a new room"),
        'title': _("New room"),
        'class': RoomResource
    }
}


def get_daypass_form(self, request):
    return get_resource_form(self, request, 'daypass')


def get_room_form(self, request):
    return get_resource_form(self, request, 'room')


def get_resource_form(self, request, type=None):
    if isinstance(self, ResourceCollection):
        assert type is not None
        model = RESOURCE_TYPES[type]['class']()
    else:
        model = self

    return model.with_content_extensions(ResourceForm, request)


@TownApp.html(model=ResourceCollection, template='resources.pt',
              permission=Public)
def view_resources(self, request):
    resources = self.query().order_by(nullsfirst(Resource.group)).all()
    resources = request.exclude_invisible(resources)
    grouped = OrderedDict()

    for group, items in groupby(resources, lambda r: r.group or _("General")):
        grouped[group] = sorted(tuple(items), key=lambda r: r.title)

    if len(grouped) == 1:
        grouped = {None: tuple(grouped.values())[0]}

    return {
        'title': _("Reservations"),
        'resources': grouped,
        'layout': ResourcesLayout(self, request)
    }


@TownApp.form(model=ResourceCollection, name='neuer-raum',
              template='form.pt', permission=Private, form=get_room_form)
def handle_new_room(self, request, form):
    return handle_new_resource(self, request, form, 'room')


@TownApp.form(model=ResourceCollection, name='neue-tageskarte',
              template='form.pt', permission=Private, form=get_daypass_form)
def handle_new_daypass(self, request, form):
    return handle_new_resource(self, request, form, 'daypass')


def handle_new_resource(self, request, form, type):
    if form.submitted(request):

        resource = self.add(
            title=form.title.data, type=type, timezone='Europe/Zurich'
        )
        form.populate_obj(resource)

        request.success(RESOURCE_TYPES[type]['success'])
        return morepath.redirect(request.link(resource))

    layout = ResourcesLayout(self, request)
    layout.include_editor()
    layout.include_code_editor()
    layout.breadcrumbs.append(Link(RESOURCE_TYPES[type]['title'], '#'))

    return {
        'layout': layout,
        'title': _(RESOURCE_TYPES[type]['title']),
        'form': form,
        'form_width': 'large'
    }


@TownApp.form(model=Resource, name='bearbeiten', template='form.pt',
              permission=Private, form=get_resource_form)
def handle_edit_resource(self, request, form):
    if form.submitted(request):
        form.populate_obj(self)

        request.success(_("Your changes were saved"))
        return morepath.redirect(request.link(self))

    elif not request.POST:
        form.process(obj=self)

    layout = ResourceLayout(self, request)
    layout.include_editor()
    layout.include_code_editor()
    layout.breadcrumbs.append(Link(_("Edit"), '#'))

    return {
        'layout': layout,
        'title': self.title,
        'form': form,
        'form_width': 'large'
    }


@TownApp.html(model=Resource, template='resource.pt', permission=Public)
def view_resource(self, request):
    return {
        'title': self.title,
        'resource': self,
        'layout': ResourceLayout(self, request),
        'feed': request.link(self, name='slots')
    }


@TownApp.view(model=Resource, request_method='DELETE', permission=Private)
def handle_delete_resource(self, request):

    if not self.deletable:
        raise exc.HTTPMethodNotAllowed()

    collection = ResourceCollection(request.app.libres_context)
    collection.delete(self, including_reservations=False)


@TownApp.form(model=Resource, permission=Private, name='cleanup',
              form=ResourceCleanupForm, template='resource_cleanup.pt')
def handle_cleanup_allocations(self, request, form):
    """ Removes all unused allocations between the given dates. """

    if form.submitted(request):
        start, end = form.data['start'], form.data['end']

        scheduler = self.get_scheduler(request.app.libres_context)
        count = scheduler.remove_unused_allocations(start, end)

        request.success(
            _("Successfully removed ${count} unused allocations", mapping={
                'count': count
            })
        )

        return morepath.redirect(request.link(self))

    layout = ResourceLayout(self, request)
    layout.breadcrumbs.append(Link(_("Clean up"), '#'))
    layout.editbar_links = None

    return {
        'layout': layout,
        'title': _("Clean up"),
        'form': form
    }


@TownApp.json(model=Resource, name='reservations', permission=Public)
def get_reservations(self, request):
    return [
        utils.ReservationInfo(reservation, request).as_dict() for reservation
        in self.bound_reservations(request)
    ]



from click.testing import CliRunner
from onegov.town.cli import cli


def test_add_town(postgres_dsn):

    runner = CliRunner()
    result = runner.invoke(cli, [
        '--dsn', postgres_dsn,
        '--schema', 'test-add-town',
        'add', 'Govikon'
    ])

    assert result.exit_code == 0
    assert 'Govikon was added to the test-add-town schema' in result.output

    result = runner.invoke(cli, [
        '--dsn', postgres_dsn,
        '--schema', 'test-add-town',
        'add', 'Govikon'
    ])

    assert result.exit_code == 1
    assert 'The schema test-add-town already contains a town' in result.output



""" Contains upgrade tasks that are executed when the application is being
upgraded on the server. See :class:`onegov.core.upgrade.upgrade_task`.

"""
from onegov.core.orm.abstract import sort_siblings
from onegov.core.upgrade import upgrade_task
from onegov.form import FormCollection
from onegov.libres import ResourceCollection
from onegov.page import PageCollection
from onegov.town import TownApp
from onegov.town.initial_content import add_builtin_forms, add_resources
from onegov.town.models.extensions import ContactExtension


@upgrade_task('Update builtin forms')
def update_builtin_forms(context, always_run=True):
    add_builtin_forms(context.session)


@upgrade_task('Add initial libres resources')
def add_initial_libres_resources(context):
    if isinstance(context.app, TownApp):
        add_resources(context.app.libres_context)


@upgrade_task('Add order to all pages')
def add_order_to_all_pages(context):
    pages = PageCollection(context.session)
    processed_siblings = set()

    for page in pages.query(ordered=False):
        siblings = page.siblings.all()

        ids = {sibling.id for sibling in siblings}
        if ids <= processed_siblings:
            continue

        sort_siblings(siblings, key=PageCollection.sort_key)

        processed_siblings.update(ids)


@upgrade_task('Update contact html')
def change_contact_html(context):
    items = []

    items.extend(PageCollection(context.session).query().all())
    items.extend(FormCollection(context.session).definitions.query().all())

    if isinstance(context.app, TownApp):
        items.extend(
            ResourceCollection(context.app.libres_context).query().all()
        )

    for item in items:
        if not isinstance(item, ContactExtension):
            continue
        if not item.contact:
            continue

        # forces a re-render of the contact html
        item.contact = item.contact



import sedate

from datetime import datetime, timedelta
from dateutil.rrule import rrule, DAILY, MO, TU, WE, TH, FR, SA, SU
from onegov.form import Form
from onegov.form.fields import MultiCheckboxField
from onegov.town import _
from wtforms.fields import RadioField
from wtforms.fields.html5 import DateField, IntegerField
from wtforms.validators import DataRequired, NumberRange, InputRequired
from wtforms_components import TimeField


WEEKDAYS = (
    (str(MO.weekday), _("Mo")),
    (str(TU.weekday), _("Tu")),
    (str(WE.weekday), _("We")),
    (str(TH.weekday), _("Th")),
    (str(FR.weekday), _("Fr")),
    (str(SA.weekday), _("Sa")),
    (str(SU.weekday), _("Su")),
)


def choices_as_integer(choices):
    if choices is None:
        return None

    return [int(c) for c in choices]


class AllocationFormHelpers(object):

    def generate_dates(self, start, end,
                       start_time=None, end_time=None, weekdays=None):
        """ Takes the given dates and generates the date tuples using rrule.
        The `except_for` field will be considered if present.

        """

        if not (start and end):
            return []

        start = start and sedate.as_datetime(start)
        end = end and sedate.as_datetime(end)

        if start == end:
            dates = (start, )
        else:
            dates = rrule(DAILY, dtstart=start, until=end, byweekday=weekdays)

        if start_time is None or end_time is None:
            return [(d, d) for d in dates]
        else:
            if end_time < start_time:
                end_offset = timedelta(days=1)
            else:
                end_offset = timedelta()

            return [
                (
                    datetime.combine(d, start_time),
                    datetime.combine(d + end_offset, end_time)
                ) for d in dates
            ]

    def combine_datetime(self, field, time_field):
        """ Takes the given date field and combines it with the given time
        field. """

        d, t = getattr(self, field).data, getattr(self, time_field).data

        if d and not t:
            return sedate.as_datetime(d)

        if not t:
            return None

        return datetime.combine(d, t)


class AllocationForm(Form, AllocationFormHelpers):
    """ Baseform for all allocation forms. Allocation forms are expected
    to implement the methods above (which contain a NotImplementedException).

    Have a look at :meth:`libres.db.scheduler.Scheduler.allocate` to find out
    more about those values.

    """

    start = DateField(
        label=_("Start"),
        validators=[InputRequired()],
        fieldset=_("Date")
    )

    end = DateField(
        label=_("End"),
        validators=[InputRequired()],
        fieldset=_("Date")
    )

    except_for = MultiCheckboxField(
        label=_("Except for"),
        choices=WEEKDAYS,
        render_kw={
            'prefix_label': False,
            'class_': 'oneline-checkboxes'
        },
        fieldset=("Date")
    )

    @property
    def weekdays(self):
        """ The rrule weekdays derived from the except_for field. """
        exceptions = {x for x in (self.except_for.data or tuple())}
        return [int(d[0]) for d in WEEKDAYS if d[0] not in exceptions]

    @property
    def dates(self):
        """ Passed to :meth:`libres.db.scheduler.Scheduler.allocate`. """
        raise NotImplementedError

    @property
    def whole_day(self):
        """ Passed to :meth:`libres.db.scheduler.Scheduler.allocate`. """
        raise NotImplementedError

    @property
    def partly_available(self):
        """  Passed to :meth:`libres.db.scheduler.Scheduler.allocate`. """
        raise NotImplementedError

    @property
    def quota(self):
        """ Passed to :meth:`libres.db.scheduler.Scheduler.allocate`. """
        raise NotImplementedError

    @property
    def quota_limit(self):
        """ Passed to :meth:`libres.db.scheduler.Scheduler.allocate`. """
        raise NotImplementedError

    @property
    def data(self):
        """ Passed to :meth:`libres.db.scheduler.Scheduler.allocate`. """
        raise NotImplementedError


class AllocationEditForm(Form, AllocationFormHelpers):
    """ Baseform for edit forms. Edit forms differ from the base allocation
    form somewhat, since they don't offer a way to generate more than
    one allocation at a time.

    The dates property is therefore expected to return a single start, end
    dates tuple.

    """

    date = DateField(
        label=_("Date"),
        validators=[InputRequired()],
        fieldset=_("Date")
    )


class Daypasses(object):

    daypasses = IntegerField(
        label=_("Daypasses"),
        validators=[
            InputRequired(),
            NumberRange(1, 999)
        ],
        fieldset=_("Daypasses")
    )

    daypasses_limit = IntegerField(
        label=_("Daypasses Limit"),
        validators=[
            InputRequired(),
            NumberRange(0, 999)
        ],
        fieldset=_("Daypasses")
    )


class DaypassAllocationForm(AllocationForm, Daypasses):

    whole_day = True
    partly_available = False
    data = None

    @property
    def quota(self):
        return self.daypasses.data

    @property
    def quota_limit(self):
        return self.daypasses_limit.data

    @property
    def dates(self):
        return self.generate_dates(
            self.start.data,
            self.end.data,
            weekdays=self.weekdays
        )


class DaypassAllocationEditForm(AllocationEditForm, Daypasses):

    whole_day = True
    partly_available = False
    data = None

    @property
    def quota(self):
        return self.daypasses.data

    @property
    def quota_limit(self):
        return self.daypasses_limit.data

    @property
    def dates(self):
        return (
            sedate.as_datetime(self.date.data),
            sedate.as_datetime(self.date.data) + timedelta(
                days=1, microseconds=-1
            )
        )

    def apply_dates(self, start, end):
        self.date.data = start.date()

    def apply_model(self, model):
        self.date.data = model.display_start().date()
        self.daypasses.data = model.quota
        self.daypasses_limit.data = model.quota_limit


class RoomAllocationForm(AllocationForm):

    as_whole_day = RadioField(
        label=_("Whole day"),
        choices=[
            ('yes', _("Yes")),
            ('no', _("No"))
        ],
        default='yes',
        fieldset=_("Date")
    )

    start_time = TimeField(
        label=_("Each starting at"),
        description=_("HH:MM"),
        validators=[InputRequired()],
        fieldset=_("Date"),
        depends_on=('as_whole_day', 'no')
    )

    end_time = TimeField(
        label=_("Each ending at"),
        description=_("HH:MM"),
        validators=[InputRequired()],
        fieldset=_("Date"),
        depends_on=('as_whole_day', 'no')
    )

    is_partly_available = RadioField(
        label=_("May be partially reserved"),
        choices=[
            ('yes', _("Yes")),
            ('no', _("No"))
        ],
        default='yes',
        fieldset=_("Options")
    )

    data = None
    quota = 1
    quota_limit = 1

    @property
    def whole_day(self):
        return self.as_whole_day.data == 'yes'

    @property
    def partly_available(self):
        return self.is_partly_available.data == 'yes'

    @property
    def dates(self):
        return self.generate_dates(
            self.start.data,
            self.end.data,
            self.start_time.data,
            self.end_time.data,
            self.weekdays
        )


class RoomAllocationEditForm(AllocationEditForm):

    as_whole_day = RadioField(
        label=_("Whole day"),
        choices=[
            ('yes', _("Yes")),
            ('no', _("No"))
        ],
        default='yes',
        fieldset=_("Date")
    )

    start_time = TimeField(
        label=_("From"),
        description=_("HH:MM"),
        validators=[DataRequired()],
        fieldset=_("Date"),
        depends_on=('as_whole_day', 'no')
    )

    end_time = TimeField(
        label=_("Until"),
        description=_("HH:MM"),
        validators=[DataRequired()],
        fieldset=_("Date"),
        depends_on=('as_whole_day', 'no')
    )

    data = None
    quota = 1
    quota_limit = 1

    @property
    def whole_day(self):
        return self.as_whole_day.data == 'yes'

    @property
    def partly_available(self):
        return self.is_partly_available.data == 'yes'

    @property
    def dates(self):
        if self.whole_day:
            return (
                sedate.as_datetime(self.date.data),
                sedate.as_datetime(self.date.data) + timedelta(
                    days=1, microseconds=-1
                )
            )
        else:
            return (
                self.combine_datetime('date', 'start_time'),
                self.combine_datetime('date', 'end_time'),
            )

    def apply_dates(self, start, end):
        self.date.data = start.date()
        self.start_time.data = start.time()
        self.end_time.data = end.time()

    def apply_model(self, model):
        self.apply_dates(model.display_start(), model.display_end())
        self.as_whole_day.data = model.whole_day and 'yes' or 'no'



""" Contains the models describing the images. """

import PIL

from dateutil.relativedelta import relativedelta
from datetime import datetime
from datetime import date
from onegov.town import _
from onegov.town.models import File
from onegov.town.models import FileCollection


class ImageCollection(FileCollection):
    """ Defines the collection of images uploaded to the site. Currently
    this is done without any ORM backing (and therefore without any
    special features like tagging, metadata and so on).

    Instead it's simply a list of images in a directory.

    This can be made more powerful (and complicated) once we have sufficent
    time to do it.

    """

    thumbnail_dimension = (256, 256)
    thumbnail_quality = 80
    allowed_mime = {'image/png', 'image/jpeg', 'image/gif', 'image/svg+xml'}
    thumbnail_extensions = {'png', 'jpg', 'jpeg', 'gif'}

    def __init__(self, app):
        assert app.has_filestorage

        self.path_prefix = 'images/'
        self.file_storage = app.filestorage.makeopendir('images')
        self.thumbnail_storage = self.file_storage.makeopendir('thumbnails')

    @property
    def files(self):
        """ Returns the :class:`~onegov.town.model.Image` instances in this
        collection.

        """

        images = self.file_storage.ilistdirinfo(files_only=True)
        images = sorted(images, key=lambda i: i[1]['modified_time'])

        for filename, info in images:
            yield Image(filename, info)

    def grouped_files(self, today=None):
        """ Returns the :class:`~onegov.town.model.Image` instances in this
        collection grouped by natural language dates.
        """
        if today is None:
            today = date.today()
        this_month = datetime(today.year, today.month, 1)
        ranges = [
            [_('This month'), this_month],
            [_('Last month'), this_month - relativedelta(months=1)],
            [_('This year'), datetime(today.year, 1, 1)],
            [_('Last year'), datetime(today.year - 1, 1, 1)],
            [_('Older'), datetime(today.year - 2, 1, 1)],
        ]
        ranges = sorted(ranges, reverse=True, key=lambda x: x[1])
        images = sorted(
            self.files, reverse=True, key=lambda x: x.info['modified_time']
        )

        # iterate over all images and append the indexes to the ranges
        last_index = 0
        range_index = 0
        for index, image in enumerate(images):
            timestamp = image.info['modified_time']
            if timestamp < ranges[range_index][1]:
                ranges[range_index].append(last_index)
                ranges[range_index].append(index)
                last_index = index

                # skip to next possible range
                range_index = len(ranges)
                for next_range in ranges:
                    if timestamp < next_range[1]:
                        range_index = ranges.index(next_range) + 1
                if range_index >= len(ranges):
                    range_index = len(ranges) - 1
                    break
        ranges[range_index].append(last_index)
        ranges[range_index].append(len(images))

        groups = []
        for range in ranges:
            if len(range) == 4 and range[2] != range[3]:
                groups.append([range[0], images[range[2]:range[3]]])
        return groups

    @property
    def thumbnails(self):
        """ Returns the :class:`~onegov.town.model.Thumbnail` instances in this
        collection.

        """
        images = self.thumbnail_storage.ilistdirinfo(files_only=True)
        images = sorted(images, key=lambda i: i[1]['modified_time'])

        for filename, info in images:
            yield Thumbnail(filename, info)

    def get_file_by_filename(self, filename):
        """ Returns the :class:`~onegov.town.model.Image` instance with the
        given name, or None if not found.

        """
        if self.file_storage.exists(filename):
            return Image(filename)

    def get_thumbnail_by_filename(self, filename):
        """ Returns the :class:`~onegov.town.model.Thumbnail` instance with the
        given name, or None if not found.

        This method *generates* the thumbnail if it doesn't exist yet!

        """
        if not self.file_storage.exists(filename):
            return None

        if self.thumbnail_storage.exists(filename):
            return Thumbnail(filename)

        extension = filename.split('.')[-1]

        if extension not in self.thumbnail_extensions:
            return Image(filename)

        im = PIL.Image.open(self.file_storage.getsyspath(filename))
        im.thumbnail(self.thumbnail_dimension)
        im.save(
            self.thumbnail_storage.open(filename, 'wb'),
            PIL.Image.EXTENSION['.' + extension],
            quality=self.thumbnail_quality
        )

        return Thumbnail(filename)

    def delete_file_by_filename(self, filename):
        """ Deletes both the image and the thumbnail of the given filename. """

        if self.file_storage.exists(filename):
            self.file_storage.remove(filename)

        if self.thumbnail_storage.exists(filename):
            self.thumbnail_storage.remove(filename)


class Image(File):
    """ A filestorage file that points to a full image (not a thumbnail). """

    @property
    def thumbnail(self):
        return Thumbnail(self.filename)

    @property
    def path(self):
        return 'images/' + self.filename


class Thumbnail(File):
    """ A filestorage file that points to a thumbnail or the original file
    storage file, if there can't be a thumbnail (say for ``*.svg``).

    Thumbnails are created on the fly and cached.

    """

    @property
    def image(self):
        return Image(self.filename)

    @property
    def path(self):
        return 'images/thumbnails/' + self.filename



""" Contains custom converters used by onegov.town. """

import morepath
from datetime import date
from time import mktime, strptime


def extended_date_decode(s):
    """ Decodes a date string HTML5 (RFC3339) compliant."""
    if not s:
        return None

    return date.fromtimestamp(mktime(strptime(s, '%Y-%m-%d')))


def extended_date_encode(d):
    """ Encodes a date HTML5 (RFC3339) compliant. """
    if not d:
        return ''

    return d.strftime('%Y-%m-%d')


extended_date_converter = morepath.Converter(
    decode=extended_date_decode, encode=extended_date_encode
)



from onegov.core.utils import ensure_scheme
from onegov.form import Form
from onegov.town import _
from wtforms import StringField, TextAreaField, validators
from wtforms.fields.html5 import EmailField


class PersonForm(Form):
    """ Form to edit people. """

    first_name = StringField(_("First name"), [validators.InputRequired()])
    last_name = StringField(_("Last name"), [validators.InputRequired()])

    function = StringField(_("Function"))

    picture_url = StringField(
        label=_("Picture"),
        description=_("URL pointing to the picture"),
        render_kw={'class_': 'image-url'}
    )

    email = EmailField(_("E-Mail"))
    phone = StringField(_("Phone"))
    website = StringField(_("Website"), filters=[ensure_scheme])

    address = TextAreaField(
        label=_("Address"),
        render_kw={'rows': 5}
    )

    notes = TextAreaField(
        label=_("Notes"),
        description=_("Public extra information about this person"),
        render_kw={'rows': 5}
    )



from datetime import date
from onegov.town.converters import extended_date_converter


def test_converters():
    converter = extended_date_converter

    assert converter.encode(None) == ['']
    assert converter.encode('') == ['']
    assert converter.encode(date(2008, 12, 30)) == ['2008-12-30']

    assert converter.decode(['']) is None
    assert converter.decode([None]) is None
    assert converter.decode(['2008-12-30']) == date(2008, 12, 30)



from onegov.town import _
from onegov.town.elements import DeleteLink, Link, LinkGroup
from onegov.town.models import Clipboard, Editor

#: Contains the messages that differ for each trait (the handling of all traits
#: is the same). New traits need to adapt the same messages as the others.
TRAIT_MESSAGES = {
    'link': dict(
        name=_("Link"),
        new_page_title=_("New Link"),
        new_page_added=_("Added a new link"),
        edit_page_title=_("Edit Link"),
        delete_message=_("The link was deleted"),
        delete_button=_("Delete link"),
        delete_question=_(
            "Do you really want to delete the link \"${title}\"?"),
    ),
    'page': dict(
        name=_("Topic"),
        new_page_title=_("New Topic"),
        new_page_added=_("Added a new topic"),
        edit_page_title=_("Edit Topic"),
        delete_message=_("The topic was deleted"),
        delete_button=_("Delete topic"),
        delete_question=_(
            "Do you really want to delete the topic \"${title}\"?"),
    ),
    'news': dict(
        name=_("News"),
        new_page_title=_("Add News"),
        new_page_added=_("Added news"),
        edit_page_title=_("Edit News"),
        delete_message=_("The news was deleted"),
        delete_button=_("Delete news"),
        delete_question=_(
            "Do you really want to delete the news \"${title}\"?"),
    )
}


class TraitInfo(object):
    """" Typically used as a mixin for adjacency list based models,
    this class provides access to the trait related methods.

    Traits are like subtypes of models. For example, the Page model has
    a page trait and a link trait. Both are managed under the same tree as
    the same type (Page), but one is rendered as a redirect, the other as
    a normal page.

    """

    @property
    def trait(self):
        """ Gets the trait of the page. """
        return self.meta.get('trait')

    @trait.setter
    def trait(self, trait):
        """ Sets the trait of the page. """
        self.meta['trait'] = trait

    @property
    def trait_messages(self):
        """ Returns all trait_messages. """
        return TRAIT_MESSAGES

    @property
    def allowed_subtraits(self):
        """ Returns a list of traits that this page may contain. """
        raise NotImplementedError

    @property
    def paste_target(self):
        """ Returns the page that should be used as parent for the content
        pasting if paste is called on the current page (self).

        This is usually just self. If the paste action should put the content
        alongside the current page, it would be the parent.

        """
        raise NotImplementedError

    def is_supported_trait(trait):
        """ Returns true if the given trait is supported by this type This
        doesn't mean that the trait may be added to this page, it serves
        as a simple sanity check, returning True if the combination of the
        type and the trait make any sense at all.

        """
        raise NotImplementedError

    def get_form_class(self, trait):
        """ Returns the form class for the given trait. """
        raise NotImplementedError

    def get_editbar_links(self, request):
        """ Returns the editbar links on the private view of this trait. """
        links = list(self.get_edit_links(request))
        links.append(
            LinkGroup(
                title=_('Add'),
                links=tuple(self.get_add_links(request))
            )
        )

        return links

    def get_add_links(self, request):
        """ Yields the add links shown on the private view of this trait. """

        for trait in self.allowed_subtraits:

            name = self.trait_messages[trait]['name']

            yield Link(
                name,
                request.link(Editor('new', self, trait)),
                classes=(
                    'new-{}'.format(trait),
                    'show-new-content-placeholder'
                )
            )

    def get_edit_links(self, request):
        """ Yields the edit links shown on the private view of this trait. """

        if self.editable:
            yield Link(
                _("Edit"),
                request.link(Editor('edit', self)),
                classes=('edit-link', )
            )
            yield Link(
                _("Copy"),
                request.link(Clipboard.from_url(request, request.path_info)),
                classes=('copy-link', )
            )

        if request.browser_session.has('clipboard_url'):

            yield Link(
                _("Paste"),
                request.link(Editor('paste', self.paste_target)),
                classes=('paste-link', 'show-new-content-placeholder'),
            )

        if self.deletable:

            trait_messages = self.trait_messages[self.trait]
            safe_delete = False if self.children else True

            if safe_delete or request.current_role == 'admin':

                if not safe_delete:
                    extra_warning = _(
                        "Please note that this page has subpages "
                        "which will also be deleted!"
                    )
                else:
                    extra_warning = ""

                yield DeleteLink(
                    _("Delete"), request.link(self),
                    confirm=_(trait_messages['delete_question'], mapping={
                        'title': self.title
                    }),
                    yes_button_text=trait_messages['delete_button'],
                    extra_information=extra_warning,
                    redirect_after=request.link(self.parent)
                )
            else:
                yield DeleteLink(
                    text=_("Delete"),
                    url=request.link(self),
                    confirm=_("This page can't be deleted."),
                    extra_information=_(
                        "This page has subpages. Only administrators can "
                        "delete pages with subpages. To delete this page, "
                        "delete all subpages first or ask an administrator "
                        "to do it for you."
                    )
                )



import os
import pytz

from datetime import datetime, date
from io import BytesIO
from onegov.core.request import CoreRequest
from onegov.core.utils import module_path, rchop
from onegov.page import PageCollection
from onegov.testing import utils
from onegov.town.models import (
    Clipboard,
    FileCollection,
    ImageCollection,
    SiteCollection
)
from unittest.mock import Mock, patch


def test_file_collection(town_app):

    collection = FileCollection(town_app)
    assert list(collection.files) == []
    assert collection.get_file_by_filename('b.txt') is None

    collection.store_file(BytesIO(b'test'), 'b.txt')
    assert len(list(collection.files)) == 1
    assert collection.get_file_by_filename('b.txt') is not None

    collection.store_file(BytesIO(b'test'), 'a.txt')
    assert len(list(collection.files)) == 2
    assert collection.get_file_by_filename('a.txt') is not None
    assert [f.filename for f in collection.files] == ['a.txt', 'b.txt']

    collection.delete_file_by_filename('a.txt')
    collection.delete_file_by_filename('b.txt')
    assert len(list(collection.files)) == 0
    assert collection.get_file_by_filename('a.txt') is None
    assert collection.get_file_by_filename('b.txt') is None


def test_image_collection(town_app):

    collection = ImageCollection(town_app)
    assert list(collection.files) == []
    assert list(collection.thumbnails) == []

    assert collection.get_file_by_filename('test.jpg') is None
    collection.store_file(utils.create_image(), 'test.jpg')
    assert collection.get_file_by_filename('test.jpg') is not None

    assert len(list(collection.files)) == 1
    assert len(list(collection.thumbnails)) == 0

    assert collection.get_thumbnail_by_filename('test.jpg') is not None

    assert len(list(collection.files)) == 1
    assert len(list(collection.thumbnails)) == 1

    collection.delete_file_by_filename('test.jpg')

    assert len(list(collection.files)) == 0
    assert len(list(collection.thumbnails)) == 0


def test_clipboard(town_app):

    request = CoreRequest(environ={
        'PATH_INFO': '/',
        'SERVER_NAME': '',
        'SERVER_PORT': '',
        'SERVER_PROTOCOL': 'https'
    }, app=town_app)

    clipboard = Clipboard.from_url(request, 'https://google.ch')
    assert clipboard.url == 'https://google.ch'

    clipboard.store_in_session()
    assert clipboard.from_session(clipboard.request).url == 'https://google.ch'

    clipboard.clear()
    assert clipboard.from_session(clipboard.request).url is None

    clipboard = Clipboard(request, clipboard.token + 'x')
    assert clipboard.url is None


def test_sitecollection(town_app):

    sitecollection = SiteCollection(town_app.session())
    objects = sitecollection.get()

    assert {o.name for o in objects['topics']} == {
        'leben-wohnen',
        'kultur-freizeit',
        'bildung-gesellschaft',
        'gewerbe-tourismus',
        'politik-verwaltung'
    }

    assert {o.name for o in objects['news']} == {
        'aktuelles',
        'willkommen-bei-onegov'
    }

    paths = (p for p in os.listdir(module_path('onegov.town', 'forms')))
    paths = (p for p in paths if p.endswith('.form'))
    paths = (os.path.basename(p) for p in paths)
    builtin_forms = set(rchop(p, '.form') for p in paths)

    assert {o.name for o in objects['forms']} == set(builtin_forms)

    assert {o.name for o in objects['resources']} == {'sbb-tageskarte'}


def test_news_years(town_app):

    collection = PageCollection(town_app.session())
    news = collection.add_root("News", type='news')
    one = collection.add(news, title="One", type='news')
    two = collection.add(news, title="Two", type='news')

    assert news.years == [datetime.utcnow().year]

    one.created = datetime(2016, 2, 1, tzinfo=pytz.utc)
    two.created = datetime(2015, 2, 1, tzinfo=pytz.utc)

    assert news.years == [2016, 2015]


def test_image_grouping(town_app):

    class MockImage(object):
        def __init__(self, mod_time):
            self.info = {'modified_time': mod_time}

    with patch('onegov.town.models.ImageCollection.files') as files:
        today = date(2008, 8, 8)

        images = [
            MockImage(datetime(2008, 9, 1, 16, 00)),  # (next month)
            MockImage(datetime(2008, 8, 3, 16, 00)),  # this month
            MockImage(datetime(2008, 8, 2, 16, 00)),  # this month
            MockImage(datetime(2008, 8, 1, 0, 00)),  # this month
            MockImage(datetime(2008, 7, 1, 16, 00)),  # last month
            MockImage(datetime(2008, 5, 1, 16, 00)),  # this year
            MockImage(datetime(2008, 2, 4, 12, 00)),  # this year
            MockImage(datetime(2007, 12, 1, 16, 00)),  # last year
            MockImage(datetime(2007, 10, 2, 14, 00)),  # last year
            MockImage(datetime(2005, 4, 1, 16, 00)),  # older
            MockImage(datetime(2002, 6, 4, 16, 00)),  # older
        ]
        files.__get__ = Mock(return_value=images)

        groups = ImageCollection(town_app).grouped_files(today=today)
        assert groups[0][0] == 'This month'
        assert groups[0][1] == images[0:4]
        assert groups[1][0] == 'Last month'
        assert groups[1][1] == images[4:5]
        assert groups[2][0] == 'This year'
        assert groups[2][1] == images[5:7]
        assert groups[3][0] == 'Last year'
        assert groups[3][1] == images[7:9]
        assert groups[4][0] == 'Older'
        assert groups[4][1] == images[9:]

    with patch('onegov.town.models.ImageCollection.files') as files:
        today = date(2008, 8, 8)

        images = [
            MockImage(datetime(2008, 8, 3, 16, 00)),  # this month
            MockImage(datetime(2008, 5, 1, 16, 00)),  # this year
            MockImage(datetime(2002, 6, 4, 16, 00)),  # older
        ]
        files.__get__ = Mock(return_value=images)

        groups = ImageCollection(town_app).grouped_files(today=today)
        assert groups[0][0] == 'This month'
        assert groups[0][1] == images[0:1]
        assert groups[1][0] == 'This year'
        assert groups[1][1] == images[1:2]
        assert groups[2][0] == 'Older'
        assert groups[2][1] == images[2:]

    with patch('onegov.town.models.ImageCollection.files') as files:
        today = date(2008, 8, 8)

        images = [
            MockImage(datetime(2008, 7, 1, 16, 00)),  # last month
            MockImage(datetime(2002, 6, 4, 16, 00)),  # older
        ]
        files.__get__ = Mock(return_value=images)

        groups = ImageCollection(town_app).grouped_files(today=today)
        assert groups[0][0] == 'Last month'
        assert groups[0][1] == images[0:1]
        assert groups[1][0] == 'Older'
        assert groups[1][1] == images[1:]

    with patch('onegov.town.models.ImageCollection.files') as files:
        today = date(2008, 8, 8)

        images = [
            MockImage(datetime(2008, 8, 1, 0, 00)),  # this month
            MockImage(datetime(2008, 2, 4, 12, 00)),  # this year
        ]
        files.__get__ = Mock(return_value=images)

        groups = ImageCollection(town_app).grouped_files(today=today)
        assert groups[0][0] == 'This month'
        assert groups[0][1] == images[0:1]
        assert groups[1][0] == 'This year'
        assert groups[1][1] == images[1:]

    with patch('onegov.town.models.ImageCollection.files') as files:
        today = date(2008, 8, 8)

        images = [
            MockImage(datetime(2008, 8, 1, 0, 00)),  # this month
        ]
        files.__get__ = Mock(return_value=images)

        groups = ImageCollection(town_app).grouped_files(today=today)
        assert groups[0][0] == 'This month'
        assert groups[0][1] == images[0:1]

    with patch('onegov.town.models.ImageCollection.files') as files:
        today = date(2008, 8, 8)

        images = [
            MockImage(datetime(2008, 7, 1, 0, 00)),  # last month
        ]
        files.__get__ = Mock(return_value=images)

        groups = ImageCollection(town_app).grouped_files(today=today)
        assert groups[0][0] == 'Last month'
        assert groups[0][1] == images[0:1]

    with patch('onegov.town.models.ImageCollection.files') as files:
        today = date(2008, 8, 8)

        images = [
            MockImage(datetime(2002, 7, 1, 0, 00)),  # older
        ]
        files.__get__ = Mock(return_value=images)

        groups = ImageCollection(town_app).grouped_files(today=today)
        assert groups[0][0] == 'Older'
        assert groups[0][1] == images[0:1]

    with patch('onegov.town.models.ImageCollection.files') as files:
        today = date(2008, 1, 8)

        images = [
            MockImage(datetime(2008, 1, 3, 16, 00)),  # this month/year
            MockImage(datetime(2008, 1, 2, 16, 00)),  # this month/year
            MockImage(datetime(2008, 1, 1, 0, 00)),  # this month/year
            MockImage(datetime(2007, 12, 14, 16, 00)),  # last month/year
            MockImage(datetime(2007, 12, 12, 16, 00)),  # last month/year
            MockImage(datetime(2007, 10, 2, 14, 00)),  # last year
            MockImage(datetime(2002, 6, 4, 16, 00)),  # older
        ]
        files.__get__ = Mock(return_value=images)

        groups = ImageCollection(town_app).grouped_files(today=today)
        assert groups[0][0] == 'This month'
        assert groups[0][1] == images[0:3]
        assert groups[1][0] == 'Last month'
        assert groups[1][1] == images[3:5]
        assert groups[2][0] == 'Last year'
        assert groups[2][1] == images[5:6]
        assert groups[3][0] == 'Older'
        assert groups[3][1] == images[6:]



""" The onegov town collection of images uploaded to the site. """
from onegov.core.filestorage import delete_static_file
from onegov.core.security import Private
from onegov.town import _
from onegov.town.app import TownApp
from onegov.town.elements import Img, Link
from onegov.town.layout import DefaultLayout
from onegov.town.models import ImageCollection, Image


@TownApp.html(model=ImageCollection, template='images.pt', permission=Private)
def view_get_image_collection(self, request):
    request.include('dropzone')

    image_groups = [
        [group, [Img(src=request.link(image.thumbnail),
                     url=request.link(image)) for image in images]]
        for group, images in self.grouped_files()
    ]

    layout = DefaultLayout(self, request)
    layout.breadcrumbs = [
        Link(_("Homepage"), layout.homepage_url),
        Link(_("Images"), request.link(self))
    ]

    return {
        'layout': layout,
        'title': _('Images'),
        'image_groups': image_groups,
    }


# XXX find a way to switch automatically:
# https://github.com/morepath/morepath/issues/70
@TownApp.json(model=ImageCollection, permission=Private, name='json')
def view_get_image_collection_json(self, request):
    return [
        {
            'group': request.translate(group),
            'images': [{'thumb': request.link(image.thumbnail),
                        'image': request.link(image)} for image in images]
        } for group, images in self.grouped_files()
    ]


@TownApp.view(
    model=Image, request_method='DELETE', permission=Private)
def delete_image(self, request):
    """ Overrides the :meth:`onegov.core.filestorage.delete_static_file` to
    ensure that thumbnails are deleted together with the image.

    """

    delete_static_file(self, request)
    ImageCollection(request.app).delete_file_by_filename(self.filename)



""" The authentication views. """

from onegov.core.security import Public, Private
from onegov.town import _
from onegov.town.app import TownApp
from onegov.town.elements import Link
from onegov.town.layout import DefaultLayout
from onegov.user import Auth
from onegov.user.forms import LoginForm


@TownApp.form(model=Auth, name='login', template='login.pt', permission=Public,
              form=LoginForm)
def handle_login(self, request, form):
    """ Handles the login requests. """

    if form.submitted(request):
        response = self.login_to(request=request, **form.login_data)

        if response:
            request.success(_("You have been logged in."))
            return response

        request.alert(_("Wrong e-mail address, password or yubikey."))

    layout = DefaultLayout(self, request)
    layout.breadcrumbs = [
        Link(_("Homepage"), layout.homepage_url),
        Link(_("Login"), request.link(self, name='login'))
    ]

    return {
        'layout': layout,
        'password_reset_link': request.link(
            request.app.town, name='request-password'),
        'title': _('Login to ${town}', mapping={
            'town': request.app.town.name
        }),
        'form': form
    }


@TownApp.html(model=Auth, name='logout', permission=Private)
def view_logout(self, request):
    """ Handles the logout requests. """

    request.info(_("You have been logged out."))
    return self.logout_to(request)



from onegov.core.layout import Layout
from onegov.form import Form
from onegov.form.fields import MultiCheckboxField
from onegov.town import _
from wtforms import StringField, TextAreaField, validators


class NewsletterForm(Form):

    title = StringField(
        label=_("Title"),
        description=_("Used in the overview and the e-mail subject"),
        validators=[validators.InputRequired()])

    lead = TextAreaField(
        label=_("Editorial"),
        description=_("A few words about this edition of the newsletter"),
        render_kw={'rows': 6})

    def update_model(self, model, request):
        model.title = self.title.data
        model.lead = self.lead.data
        model.html = self.get_html(request)

    def apply_model(self, model):
        self.title.data = model.title
        self.lead.data = model.lead

    def get_html(self, request):
        return ''

    @classmethod
    def with_news(cls, request, news):

        layout = Layout(None, request)

        choices = tuple((
            str(item.id),
            '<div class="title">{}</div><div class="date">{}</div>'.format(
                item.title,
                layout.format_date(item.created, 'relative')
            )
        ) for item in news)

        class NewsletterWithNewsForm(cls):

            news = MultiCheckboxField(
                label=_("Latest news"),
                choices=choices,
                render_kw={
                    'prefix_label': False,
                    'class_': 'recommended'
                },
                fieldset=_("Selected News / Events")
            )

            def update_model(self, model, request):
                super().update_model(model, request)
                model.content['news'] = self.news.data

            def apply_model(self, model):
                super().apply_model(model)
                self.news.data = model.content.get('news')

        return NewsletterWithNewsForm

    @classmethod
    def with_occurrences(cls, request, occurrences):

        choices = tuple(
            (
                str(item.id),
                '<div class="title">{}</div><div class="date">{}</div>'.format(
                    item.title, item.localized_start.strftime('%d.%m.%Y %H:%M')
                )
            ) for item in occurrences
        )

        class NewsletterWithOccurrencesForm(cls):

            occurrences = MultiCheckboxField(
                label=_("Events"),
                choices=choices,
                render_kw={
                    'prefix_label': False,
                    'class_': 'recommended'
                },
                fieldset=_("Selected News / Events")
            )

            def update_model(self, model, request):
                super().update_model(model, request)
                model.content['occurrences'] = self.occurrences.data

            def apply_model(self, model):
                super().apply_model(model)
                self.occurrences.data = model.content.get('occurrences')

        return NewsletterWithOccurrencesForm


class NewsletterSendForm(Form):

    @classmethod
    def for_newsletter(cls, newsletter, recipients):
        choices = tuple(
            (
                (recipient.id.hex, recipient.address)
                for recipient in recipients
                if recipient.confirmed and
                recipient not in newsletter.recipients
            )
        )

        class NewsletterSendFormWithRecipients(cls):

            recipients = MultiCheckboxField(
                label=_("Recipients"),
                choices=choices,
                default=tuple(c[0] for c in choices),
                render_kw={
                    'prefix_label': False
                },
            )

            def validate_recipients(self, field):
                if len(field.data) == 0:
                    raise validators.ValidationError(
                        _("Please select at least one recipient")
                    )

        return NewsletterSendFormWithRecipients
